// Version information for the "BasicUsageEnvironment" library
// Copyright (c) 1996-2011 Live Networks, Inc.  All rights reserved.

#ifndef _BASICUSAGEENVIRONMENT_VERSION_HH
#define _BASICUSAGEENVIRONMENT_VERSION_HH

#define BASICUSAGEENVIRONMENT_LIBRARY_VERSION_STRING	"2011.01.24"
#define BASICUSAGEENVIRONMENT_LIBRARY_VERSION_INT		1295827200

#endif
