// Version information for the "liveMedia" library
// Copyright (c) 1996-2011 Live Networks, Inc.  All rights reserved.

#ifndef _LIVEMEDIA_VERSION_HH
#define _LIVEMEDIA_VERSION_HH

#define LIVEMEDIA_LIBRARY_VERSION_STRING	"2011.01.24"
#define LIVEMEDIA_LIBRARY_VERSION_INT		1295827200

#endif
